from dataclasses import dataclass
from typing import Any
import os
import json5


@dataclass()
class RepoConfig(object):
    name: str
    path: str
    # defaults to master if not set
    default_branch: str | None = None


@dataclass()
class StatConfig(object):
    name: str
    description: str
    type: str
    params: Any
    path_in_repo: str


REPO_STORAGE_LOCATION = "repo_storage_location"


def get_default_config():
    return {
        REPO_STORAGE_LOCATION: "./repos",
        "stat_storage": {
            "type": "csv",
            "path": "./stats",  # will store stats in ./stats/<repo_name>/<stat_name>.csv
        },
        "repos": {},
    }


def get_config_path():
    return os.environ.get("REPOTRACER_CONFIG_PATH", "./config.json")


default_config = get_default_config()
config_file_contents = None


def read_config_file():
    global config_file_contents
    if config_file_contents:
        return default_config | (config_file_contents or {})
    # print("Using default config.")
    try:
        print("Looking for config file at", os.path.abspath(get_config_path()))
        with open(get_config_path()) as f:
            config_file_contents = json5.load(f)  # python 3.9 operator for dict update
            return default_config | (config_file_contents or {})
    except FileNotFoundError:
        print(f"Could not find config file at {get_config_path()}. Using defaults.")
        config_file_contents = default_config
        return config_file_contents


def get_repo_storage_location():
    return read_config_file().get(REPO_STORAGE_LOCATION, "./repos")


def get_stat_storage_config():
    return read_config_file()["stat_storage"]


def list_repos():
    try:
        return list(read_config_file()["repos"].keys())
    except KeyError:
        return []


def list_stats_for_repo(repo_name):
    try:
        return list(read_config_file()["repos"][repo_name]["stats"].keys())
    except KeyError:
        return []


def get_config(repo_name, stat_name) -> (RepoConfig, str):
    config_data = read_config_file()
    try:
        repo_subtree = config_data["repos"][repo_name]
        repo_config = RepoConfig(
            name=repo_name,
            path=repo_subtree.get("path"),
            default_branch=repo_subtree.get("default_branch"),
        )
    except KeyError:
        known_repos = ",".join(config_data["repos"].keys())
        raise Exception(
            f"Repo '{repo_name}' not found in config. Known repos are '{known_repos}'"
        )

    try:
        stat_config = repo_subtree["stats"][stat_name]
        stat_config["name"] = stat_name
    except KeyError:
        valid_stats = ", ".join(repo_subtree["stats"].keys())
        raise Exception(
            f"The stat '{stat_name}' does not exist in the config for the repo '{repo_name}'. Here are the known stats: '{valid_stats}'"
        )

    return repo_config, stat_config


def write_config_file(config):
    global config_file_contents
    with open(get_config_path(), "w") as f:
        json5.dump(config, f, indent=4, quote_keys=True)
    config_file_contents = config


def remove_nones(d: dict[Any, Any]):
    return {k: v for k, v in d.items() if v is not None}


def add_repo(repo_config: RepoConfig):
    config = read_config_file()
    config["repos"][repo_config.name] = remove_nones(
        {
            "path": repo_config.path,
            "default_branch": repo_config.default_branch
            if repo_config.default_branch != "master"
            else None,
            "stats": {},
        }
    )
    write_config_file(config)


def add_stat(repo_name: str, stat_config: StatConfig):
    config = read_config_file()
    repo_config = config["repos"][repo_name]
    if "stats" not in repo_config:
        repo_config["stats"] = {}
    repo_config["stats"][stat_config.name] = remove_nones(
        {
            "description": stat_config.description,
            "type": stat_config.type,
            "params": stat_config.params,
            "path_in_repo": stat_config.path_in_repo,
        }
    )

    write_config_file(config)



from datetime import datetime
import sh
import os
import sys
import logging
import subprocess
from .config import RepoConfig
from . import config
from rich.console import Console
from typing import Optional
import pandas as pd

# logging.basicConfig(level=logging.INFO)


def check():
    revparse = git("rev-parse", "--show-toplevel").strip()
    if revparse.endswith("repotracer"):
        print("raising")
        raise Exception(
            "(bug in repotracer): cannot run a git command against the repotracer repo itself"
        )


repodir = "/users/amedee/workspace/samplerepos/react"
git = sh.git.bake(no_pager=True)
git_normal = sh.git.bake(_out=sys.stdout, _err=sys.stderr)


def list_commits(start, end):  # -> [string, string]:
    start = start or first_commit_date()
    end = end or datetime.now().strftime("%Y-%m-%d")
    data = []
    for line in git.log(
        format="%H,%cd",
        date="iso-strict",
        since=start,
        until=end,
        no_merges=True,
        _iter=True,
    ):
        data.append(line.split(","))
    return data


def build_commit_df(
    start: Optional[str], end: Optional[str], freq: str = "D"
) -> pd.DataFrame:
    commits = pd.DataFrame(list_commits(start, end), columns=["sha", "created_at"])
    commits.created_at = pd.DatetimeIndex(
        data=pd.to_datetime(commits.created_at, utc=True)
    )
    commits = commits.set_index(
        commits.created_at,
        drop=False,
    )
    commits = commits.groupby(
        pd.Grouper(key="created_at", freq=freq)
    ).last()
    return commits


def first_commit_date():
    check()
    # from https://stackoverflow.com/a/5189296
    first_sha = git("rev-list", "--max-parents=0", "HEAD").strip()
    return git.log(first_sha, "--pretty=format:%cd", "--date=format:%Y-%m-%d")


def checkout(sha):
    check()
    return git.checkout(sha)


def reset_hard(target="HEAD"):
    check()
    return git.reset("--hard", target)


def clean_untracked():
    check()
    return git.clean("-fxd")


def current_message():
    check()
    return git.log("-1", "--pretty=%B")


def current_files():
    check()
    return git.diff("--name-only", "HEAD~")


def current_date():
    check()
    return git.log("-1", "--pretty=format:%cd")


def pull(obj="master"):
    check()
    return git.pull("origin", obj)


def get_commit_author():
    check()
    return git.log("-1", "--pretty=format:%aE")


def is_repo_setup(repo_path):
    return os.path.exists(os.path.join(repo_path, ".git"))


def get_default_branch():
    res = git("symbolic-ref", "refs/remotes/origin/HEAD")
    # return the basename of res
    return os.path.basename(res.strip())


def download_repo(url, name=None, branch=None):
    cwd = os.getcwd()
    repos_location = config.get_repo_storage_location()
    repo_name = name or os.path.splitext(os.path.basename(url))[0]
    repo_storage_path = os.path.join(repos_location, repo_name)
    os.makedirs(repo_storage_path, exist_ok=True)
    print(f"Downloading {repo_name} from {url} to {repo_storage_path}")
    os.chdir(repos_location)

    # This is a  "blobless" clone, which downloads the data for the latest commit (HEAD) and then for past commits
    # downloads only the metadata (commit summaries and then trees which are folder names + pointers to blobs which contain file contents)
    # It's useful to us because we are going to be checking out only a fraction of the commits, so
    # we won't need the file contents for the rest of the commits.
    # git_normal.clone(url, ".", "--filter=blob:none", _out=sys.stdout, _err=sys.stderr)
    # todo decide if we actually want a blobless ^

    if not branch:
        git_normal.clone(url, repo_name, "--single-branch")
    else:
        git_normal.clone(
            url,
            repo_name,
            "--single-branch",
            f"--branch={branch}",
        )

    os.chdir(repo_name)
    # after we do the blobless clone we check out the last commit of every day, to make git
    # download the blobs for those commits, so that way the first stat run won't have to do it.
    # This is technically part of downloading the repo
    # for commit in list_commits():
    #     sha = commit[0]
    #     git.checkout(sha)

    default_branch = branch or get_default_branch()
    os.chdir(cwd)
    return RepoConfig(name=repo_name, path=repo_name, default_branch=default_branch)
